"""This module contains the base class of all playlist providers."""
import logging
import time
from typing import Optional, Type, List

from django.conf import settings
from django.db import transaction
from django.db.models.expressions import F

import core.settings.storage as storage
from core.models import (
    ArchivedPlaylist,
    PlaylistEntry,
    ArchivedPlaylistQuery,
    RequestLog,
)
from core.musiq import song_utils
from core.musiq.music_provider import MusicProvider, ProviderError
from core.musiq.song_provider import SongProvider


class PlaylistProvider(MusicProvider):
    """The base class for playlist providers."""

    @staticmethod
    def create(
        query: Optional[str] = None, key: Optional[int] = None
    ) -> "PlaylistProvider":
        """Factory method to create a playlist provider.
        Both query and key need to be specified.
        Detects the type of provider needed and returns one of corresponding type."""
        if query is None:
            logging.error(
                "archived playlist requested but no query given (key %s)", key
            )
            raise ValueError
        if key is None:
            logging.error("archived playlist requested but no key given")
            raise ValueError
        try:
            archived_playlist = ArchivedPlaylist.objects.get(id=key)
        except ArchivedPlaylist.DoesNotExist:
            logging.error("archived song requested for nonexistent key %s", key)
            raise ValueError

        playlist_type = song_utils.determine_playlist_type(archived_playlist)
        provider_class: Optional[Type[PlaylistProvider]] = None
        if playlist_type == "local":
            from core.musiq.localdrive import LocalPlaylistProvider

            provider_class = LocalPlaylistProvider
        elif storage.get("youtube_enabled") and playlist_type == "youtube":
            from core.musiq.youtube import YoutubePlaylistProvider

            provider_class = YoutubePlaylistProvider
        elif storage.get("spotify_enabled") and playlist_type == "spotify":
            from core.musiq.spotify import SpotifyPlaylistProvider

            provider_class = SpotifyPlaylistProvider
        elif storage.get("soundcloud_enabled") and playlist_type == "soundcloud":
            from core.musiq.soundcloud import SoundcloudPlaylistProvider

            provider_class = SoundcloudPlaylistProvider
        elif storage.get("jamendo_enabled") and playlist_type == "jamendo":
            from core.musiq.jamendo import JamendoPlaylistProvider

            provider_class = JamendoPlaylistProvider
        elif playlist_type == "playlog":
            # The playlist may contain various song types, but all of them will be archived.
            # We can use the local playlist provider to enqueue them.
            from core.musiq.localdrive import LocalPlaylistProvider

            provider_class = LocalPlaylistProvider
        if not provider_class:
            raise NotImplementedError(f"No provider for given playlist: {query}, {key}")
        provider = provider_class(query, key)
        return provider

    @staticmethod
    def get_id_from_external_url(url: str) -> Optional[str]:
        """Constructs and returns the external id based on the given url."""
        raise NotImplementedError()

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        super().__init__(query, key)
        self.ok_message = "queueing playlist"
        self.title: Optional[str] = None
        self.urls: List[str] = []

    def check_cached(self) -> bool:
        if self.key is not None:
            archived_playlist = ArchivedPlaylist.objects.get(id=self.key)
        else:
            assert self.query is not None
            try:
                list_id = self.get_id_from_external_url(self.query)
                archived_playlist = ArchivedPlaylist.objects.get(list_id=list_id)
            except (KeyError, ArchivedPlaylist.DoesNotExist):
                return False
        self.id = archived_playlist.list_id
        self.title = archived_playlist.title
        self.key = archived_playlist.id
        self.urls = [entry.url for entry in archived_playlist.entries.all()]
        return True

    def search_id(self) -> Optional[str]:
        """Fetches the id of this playlist from the internet and returns it."""
        raise NotImplementedError()

    def check_available(self) -> bool:
        if self.id is not None:
            return True
        assert self.query
        list_id = self.get_id_from_external_url(self.query)
        if list_id is None:
            list_id = self.search_id()
        if list_id is None:
            return False
        self.id = list_id
        return True

    def is_radio(self) -> bool:
        """Returns whether this playlist is a radio.
        A radio as a playlist that was created for a given song.
        The result can be different if called another time for the same song."""
        raise NotImplementedError()

    def fetch_metadata(self) -> bool:
        """Fetches the title and list of songs for this playlist from the internet."""
        raise NotImplementedError()

    def enqueue_placeholder(self, manually_requested) -> None:
        # Playlists have no placeholder representation.
        pass

    def remove_placeholder(self) -> None:
        pass

    def make_available(self) -> bool:
        queryset = ArchivedPlaylist.objects.filter(list_id=self.id)
        if not self.is_radio() and queryset.exists():
            archived_playlist = queryset.get()
            self.key = archived_playlist.id
            self.urls = [entry.url for entry in archived_playlist.entries.all()]
        else:
            if not self.fetch_metadata():
                return False
        return True

    def persist(self, request_ip: str, archive: bool = True) -> None:
        if self.is_radio():
            return

        assert self.id
        if self.title is None:
            logging.warning("Persisting a playlist with no title (id %s)", self.id)
            self.title = ""

        with transaction.atomic():
            queryset = ArchivedPlaylist.objects.filter(list_id=self.id)
            if queryset.count() == 0:
                initial_counter = 1 if archive else 0
                archived_playlist = ArchivedPlaylist.objects.create(
                    list_id=self.id, title=self.title, counter=initial_counter
                )
                for index, url in enumerate(self.urls):
                    PlaylistEntry.objects.create(
                        playlist=archived_playlist, index=index, url=url
                    )
            else:
                if archive:
                    queryset.update(counter=F("counter") + 1)
                archived_playlist = queryset.get()

        if archive:
            ArchivedPlaylistQuery.objects.get_or_create(
                playlist=archived_playlist, query=self.query
            )

        if storage.get("logging_enabled") and request_ip:
            RequestLog.objects.create(playlist=archived_playlist, address=request_ip)

    def enqueue(self) -> None:
        for index, external_url in enumerate(self.urls):
            if index == storage.get("max_playlist_items"):
                break
            # request every url in the playlist as their own url
            try:
                song_provider = SongProvider.create(external_url=external_url)
                song_provider.request("", archive=False, manually_requested=False)
            except (ProviderError, NotImplementedError) as e:
                logging.warning(
                    "Error while enqueuing url %s to playlist %s: %s",
                    external_url,
                    self.title,
                    self.id,
                )
                logging.exception(e)
                continue

            if settings.DEBUG:
                # the sqlite database has problems if songs are pushed very fast
                # while a new song is taken from the queue. Add a delay to mitigate.
                time.sleep(1)



"""This module contains all Spotify related code."""

from __future__ import annotations

from typing import Optional, List, Tuple, TYPE_CHECKING
from urllib.parse import urlparse

from django.http.response import HttpResponse

import core.settings.storage as storage
from core.musiq import song_utils, musiq
from core.musiq.song_provider import SongProvider
from core.musiq.playlist_provider import PlaylistProvider
from core.musiq.spotify_web import OAuthClient

if TYPE_CHECKING:
    from core.musiq.song_utils import Metadata


class Spotify:
    """This class contains code for both the song and playlist provider"""

    _web_client: OAuthClient = None  # type: ignore

    @property
    def web_client(self) -> OAuthClient:
        """Returns the web client if it was already created.
        If not, it is created using the spotify credentials from the database."""
        if Spotify._web_client is None:
            client_id = storage.get(key="spotify_client_id")
            client_secret = storage.get(key="spotify_client_secret")
            Spotify._web_client = OAuthClient(
                base_url="https://api.spotify.com/v1",
                refresh_url="https://auth.mopidy.com/spotify/token",
                client_id=client_id,
                client_secret=client_secret,
            )
        return Spotify._web_client

    def get_search_suggestions(
        self, query: str, playlist: bool
    ) -> List[Tuple[str, str]]:
        """Returns a list of suggested items for the given query.
        Returns playlists if :param playlist: is True, songs otherwise."""
        result = self.web_client.get(
            "search",
            params={
                "q": query,
                "limit": "20",
                "market": "from_token",
                "type": "playlist" if playlist else "track",
            },
        )

        if playlist:
            items = result["playlists"]["items"]
        else:
            items = result["tracks"]["items"]

        suggestions = []
        for item in items:
            external_url = item["external_urls"]["spotify"]
            title = item["name"]
            if playlist:
                displayname = title
            else:
                artist = item["artists"][0]["name"]
                # apply filter from the settings
                if song_utils.is_forbidden(artist) or song_utils.is_forbidden(title):
                    continue
                displayname = song_utils.displayname(artist, title)
            suggestions.append((displayname, external_url))

        # remove duplicates and filter by keywords
        chosen_displaynames = set()
        unique_suggestions = []
        for suggestion in suggestions:
            if suggestion[0] in chosen_displaynames:
                continue
            unique_suggestions.append(suggestion)
            chosen_displaynames.add(suggestion[0])
        return unique_suggestions


class SpotifySongProvider(SongProvider, Spotify):
    """This class handles songs from Spotify."""

    @staticmethod
    def get_id_from_external_url(url: str) -> str:
        return urlparse(url).path.split("/")[-1]

    @staticmethod
    def get_id_from_internal_url(url: str) -> str:
        """Returns the internal id based on the given url."""
        return url.split(":")[-1]

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "spotify"
        super().__init__(query, key)

        self.metadata: "Metadata" = {}

    def check_cached(self) -> bool:
        # Spotify songs cannot be cached and have to be streamed everytime
        return False

    def check_available(self) -> bool:
        if not self.gather_metadata():
            return False
        # the default bitrate of mopidy-spotify is 160kbps
        # estimate the size of a song by multiplying with its duration
        size = self.metadata["duration"] * 160 / 8 * 1000
        return self.check_not_too_large(size)

    def gather_metadata(self) -> bool:
        """Fetches metadata for this song's uri from Spotify."""
        if not self.id:
            results = self.web_client.get(
                "search",
                params={
                    "q": self.query,
                    "limit": "50",
                    "market": "from_token",
                    "type": "track",
                },
            )

            # apply the filterlist from the settings
            for item in results["tracks"]["items"]:
                artist = item["artists"][0]["name"]
                title = item["name"]
                if song_utils.is_forbidden(artist) or song_utils.is_forbidden(title):
                    continue
                result = item
                break
            else:
                # all tracks got filtered
                self.error = "All results filtered"
                return False
            self.id = result["id"]
        else:
            result = self.web_client.get(f"tracks/{self.id}", params={"limit": "1"})
        try:
            self.metadata["artist"] = result["artists"][0]["name"]
            self.metadata["title"] = result["name"]
            self.metadata["duration"] = result["duration_ms"] / 1000
            self.metadata["internal_url"] = result["uri"]
            self.metadata["external_url"] = result["external_urls"]["spotify"]
            self.metadata["stream_url"] = None
        except KeyError:
            self.error = "No song found"
            return False
        return True

    def get_metadata(self) -> "Metadata":
        if not self.metadata:
            self.gather_metadata()
        return self.metadata

    def _get_path(self) -> str:
        # spotify is not cached in the cache directory
        raise NotImplementedError()

    def get_internal_url(self) -> str:
        if not self.id:
            raise ValueError()
        return "spotify:track:" + self.id

    def get_external_url(self) -> str:
        if not self.id:
            raise ValueError()
        return "https://open.spotify.com/track/" + self.id

    def get_suggestion(self) -> str:
        result = self.web_client.get(
            "recommendations",
            params={"limit": "1", "market": "from_token", "seed_tracks": self.id},
        )

        try:
            external_url = result["tracks"][0]["external_urls"]["spotify"]
        except IndexError:
            self.error = "no recommendation found"
            raise ValueError("No suggested track")

        return external_url

    def request_radio(self, request_ip: str) -> HttpResponse:
        result = self.web_client.get(
            "recommendations",
            params={
                "limit": storage.get("max_playlist_items"),
                "market": "from_token",
                "seed_tracks": self.id,
            },
        )

        for track in result["tracks"]:
            external_url = track["external_urls"]["spotify"]
            musiq.do_request_music(
                "",
                external_url,
                None,
                False,
                "spotify",
                archive=False,
                manually_requested=False,
            )

        return HttpResponse("queueing radio")


class SpotifyPlaylistProvider(PlaylistProvider, Spotify):
    """This class handles Spotify Playlists."""

    @staticmethod
    def get_id_from_external_url(url: str) -> Optional[str]:
        if not (
            url.startswith("https://open.spotify.com/playlist/")
            or url.startswith("https://open.spotify.com/artist/")
            or url.startswith("https://open.spotify.com/album/")
        ):
            return None
        return urlparse(url).path.split("/")[-1]

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "spotify"
        # can be one of playlists, artists or albums
        # defaults to playlist, only changes if an artist or album is found during search_id
        # this type is not reflected in list_id. Thus, ArchivedPlaylist entries do not know
        # what kind of collection of songs they are.
        # This is considered acceptable, generating external urls from playlist is never required
        # and finding cached lists still works as extracted ids still match
        self._spotify_endpoint = "playlists"
        if query:
            if query.startswith("https://open.spotify.com/playlist/"):
                self._spotify_endpoint = "playlists"
            elif query.startswith("https://open.spotify.com/artist/"):
                self._spotify_endpoint = "artists"
            elif query.startswith("https://open.spotify.com/album/"):
                self._spotify_endpoint = "albums"
            super().__init__(query, key)

    def search_id(self) -> Optional[str]:
        result = self.web_client.get(
            "search",
            params={
                "q": self.query,
                "limit": "1",
                "market": "from_token",
                "type": "artist,album,playlist",
            },
        )

        try:
            list_info = result["artists"]["items"][0]
            self._spotify_endpoint = "artists"
        except IndexError:
            try:
                list_info = result["albums"]["items"][0]
                self._spotify_endpoint = "albums"
            except IndexError:
                try:
                    list_info = result["playlists"]["items"][0]
                    self._spotify_endpoint = "playlists"
                except IndexError:
                    self.error = "No playlist found"
                    return None

        list_id = list_info["id"]
        self.title = list_info["name"]

        return list_id

    def is_radio(self) -> bool:
        return False

    def fetch_metadata(self) -> bool:
        if self.title is None:
            result = self.web_client.get(
                f"{self._spotify_endpoint}/{self.id}", params={"fields": "name"}
            )
            self.title = result["name"]

        # download at most 50 tracks for a playlist (spotifys maximum)
        # for more tracks paging would need to be implemented
        if self._spotify_endpoint == "playlists":
            result = self.web_client.get(
                f"playlists/{self.id}/tracks",
                params={
                    "fields": "items(track(external_urls(spotify)))",
                    "limit": "50",
                    "market": "from_token",
                },
            )
            track_infos = result["items"]
            for track_info in track_infos:
                self.urls.append(track_info["track"]["external_urls"]["spotify"])
        elif self._spotify_endpoint == "artists":
            result = self.web_client.get(
                f"artists/{self.id}/top-tracks",
                params={
                    "fields": "tracks(external_urls(spotify))",
                    "limit": "50",
                    "market": "from_token",
                },
            )
            tracks = result["tracks"]
            for track in tracks:
                self.urls.append(track["external_urls"]["spotify"])
        elif self._spotify_endpoint == "albums":
            result = self.web_client.get(
                f"albums/{self.id}/tracks",
                params={
                    "fields": "items(external_urls(spotify))",
                    "limit": "50",
                    "market": "from_token",
                },
            )
            tracks = result["items"]
            for track in tracks:
                self.urls.append(track["external_urls"]["spotify"])

        return True



"""This module contains the worker class calculating the visualizations."""

from __future__ import annotations

from threading import Thread, Event
import time
from typing import Dict, TYPE_CHECKING, TypeVar

from django.db import connection

from core import redis
from core.celery import app
from core.lights.circle.circle import Circle
from core.lights.device import Device
from core.lights.programs import Adaptive, LedProgram, ScreenProgram, VizProgram
from core.lights.programs import Alarm
from core.lights.programs import Cava
from core.lights.programs import Disabled
from core.lights.programs import Fixed
from core.lights.programs import Rainbow
from core.lights.exceptions import RenderingStoppedException
from core.settings import storage


if TYPE_CHECKING:
    from core.lights.ring import Ring
    from core.lights.wled import WLED
    from core.lights.strip import Strip
    from core.lights.screen import Screen

    T = TypeVar("T", Ring, WLED, Strip)  # pylint: disable=invalid-name

lights_lock = redis.lock("lights_lock")


def start() -> None:
    _loop.delay()
    connection.close()


class DeviceManager:
    """A class managing all visualization devices.
    This class maintains state, but only in the worker thread that updates the lights.
    This keeps the necessary variables local to the thread, avoiding db/redis queries each frame."""

    def __init__(self) -> None:
        from core.lights.ring import Ring
        from core.lights.wled import WLED
        from core.lights.strip import Strip
        from core.lights.screen import Screen

        self.ups = 30
        self.seconds_per_frame = 1 / self.ups

        self.loop_active = Event()

        self.disabled_program = Disabled(self)

        self.ring = Ring(self)
        self.wled = WLED(self)
        self.strip = Strip(self)
        self.screen = Screen(self)

        self.cava_program = Cava(self)
        self.alarm_program = Alarm(self)

        # a dictionary containing all devices by their name
        self.all_devices: Dict[str, Device] = {
            device.name: device
            for device in [self.ring, self.wled, self.strip, self.screen]
        }
        # a dictionary containing all led programs by their name
        self.led_programs: Dict[str, LedProgram] = {"Disabled": self.disabled_program}
        for led_program_class in [Fixed, Rainbow, Adaptive]:
            led_instance = led_program_class(self)
            self.led_programs[led_instance.name] = led_instance
        # a dictionary containing all screen programs by their name
        self.screen_programs: Dict[str, ScreenProgram] = {
            "Disabled": self.disabled_program
        }
        for screen_program_class in [Circle]:
            screen_instance = screen_program_class(self)
            self.screen_programs[screen_instance.name] = screen_instance
        # a dictionary containing *all* programs by their name
        self.all_programs: Dict[str, VizProgram] = {
            **self.led_programs,
            **self.screen_programs,
        }
        self.program_speed = storage.get("program_speed")
        self.fixed_color = storage.get("fixed_color")
        self.last_fixed_color = self.fixed_color

        for device in [self.ring, self.wled, self.strip, self.screen]:
            device.load_program()

        self.consumers_changed()

        self.listener = Thread(target=self.listen_for_changes)
        self.listener.start()

    def listen_for_changes(self) -> None:
        p = redis.pubsub(ignore_subscribe_messages=True)
        p.subscribe("lights_settings_changed")
        for message in p.listen():
            settings_changed = message["data"]

            if settings_changed == "stop":
                # delete the lock the main thread is checking each loop in order to stop it
                # this removes the need for an extra redis variable
                # that would need to be checked every loop
                self.loop_active.set()
                self.loop_active = None
                break

            if settings_changed == "alarm_started":
                self.alarm_started()
                continue
            if settings_changed == "alarm_stopped":
                self.alarm_stopped()
                continue

            if settings_changed == "base":
                # base settings affecting every device were modified
                self.fixed_color = storage.get("fixed_color")
                self.program_speed = storage.get("program_speed")
                continue

            if settings_changed == "adjust_screen":
                self.screen.adjust()
                continue

            # a device was changed
            # reload all settings for this device from the database
            # instead of communicating the exact changed setting.
            device_name = settings_changed
            device = self.all_devices[device_name]
            device.brightness = storage.get(f"{device_name}_brightness")
            device.monochrome = storage.get(f"{device_name}_monochrome")
            if device_name == "wled":
                self.wled.ip = storage.get("wled_ip")
                self.wled.port = storage.get("wled_port")
            program = self.all_programs[storage.get(f"{device_name}_program")]
            self.set_program(device, program)
        connection.close()

    def set_program(self, device: Device, program: VizProgram) -> None:
        # don't allow program change on disconnected devices
        if not device.initialized:
            return

        with lights_lock:
            if device.program == program:
                # nothing to do
                return

            device.program.release()
            program.use()

            device.program = program
            self.consumers_changed()

            if program.name == "Disabled":
                device.clear()

    def consumers_changed(self) -> None:
        """Stops the loop if no led is active, starts it otherwise"""
        if self.disabled_program.consumers == 4:
            self.loop_active.clear()
            redis.set("lights_active", False)
        else:
            self.loop_active.set()
            redis.set("lights_active", True)

    def alarm_started(self) -> None:
        """Makes alarm the current program but doesn't update the database."""
        self.alarm_program.use()
        self.last_fixed_color = self.fixed_color
        for device in [self.ring, self.wled, self.strip]:
            self.set_program(device, self.led_programs["Fixed"])
        # the screen program adapts with the alarm and is not changed

    def alarm_stopped(self) -> None:
        """Restores the state from before the alarm."""
        self.alarm_program.release()
        self.fixed_color = self.last_fixed_color

        # the database still contains the program from before the alarm. restore it.
        for device in [self.ring, self.wled, self.strip]:
            self.set_program(
                device, self.led_programs[storage.get(f"{device.name}_program")]
            )

    def loop(self):
        iteration_count = 0
        adaptive_quality_window = self.ups * 10
        time_sum = 0.0
        while True:
            try:
                self.loop_active.wait()
            except AttributeError:
                # the lock was deleted by the listener thread in order to stop the main thread
                # The display must be stopped from this thread, otherwise no new one can be created
                self.screen.program.stop()
                self.listener.join()
                break

            computation_start = time.time()

            with lights_lock:
                # these programs only actually do work if their respective programs are active
                self.cava_program.compute()
                self.alarm_program.compute()

                if self.screen.program.name != "Disabled":
                    try:
                        self.screen.program.draw()
                    except RenderingStoppedException:
                        self.set_program(self.screen, self.disabled_program)

                self.ring.program.compute()
                if self.wled.program != self.ring.program:
                    self.wled.program.compute()
                if self.strip.program != self.ring.program:
                    self.strip.program.compute()

                if self.ring.program.name != "Disabled":
                    if self.ring.monochrome:
                        ring_colors = [
                            self.ring.program.strip_color()
                            for _ in range(self.ring.LED_COUNT)
                        ]
                    else:
                        ring_colors = self.ring.program.ring_colors()
                    self.ring.set_colors(ring_colors)

                if self.wled.program.name != "Disabled":
                    if self.wled.monochrome:
                        wled_colors = [
                            self.wled.program.strip_color()
                            for _ in range(self.wled.led_count)
                        ]
                    else:
                        wled_colors = self.wled.program.wled_colors()
                    self.wled.set_colors(wled_colors)

                if self.strip.program.name != "Disabled":
                    strip_color = self.strip.program.strip_color()
                    self.strip.set_color(strip_color)

            computation_time = time.time() - computation_start

            if self.screen.program.name != "Disabled":
                time_sum += computation_time
                iteration_count += 1
                if (
                    iteration_count >= adaptive_quality_window
                    or time_sum
                    >= 1.5 * adaptive_quality_window * self.seconds_per_frame
                ):
                    average_computation_time = time_sum / adaptive_quality_window
                    iteration_count = 0
                    time_sum = 0.0

                    # print(f"avg: {average_computation_time/seconds_per_frame}")
                    if average_computation_time > 0.9 * self.seconds_per_frame:
                        # if the loop takes too long and a screen program is active,
                        # it can be reduced in resolution to save time
                        self.screen.program.decrease_resolution()
                    elif average_computation_time < 0.6 * self.seconds_per_frame:
                        # if the loop has time to spare and a screen program is active,
                        # we can increase its quality
                        self.screen.program.increase_resolution()

            # print(f'computation took {computation_time:.5f}s')
            try:
                time.sleep(self.seconds_per_frame - computation_time)
            except ValueError:
                pass


@app.task
def _loop() -> None:
    manager = DeviceManager()
    manager.loop()
    connection.close()



"""This module contains all Soundcloud related code."""

from __future__ import annotations

from typing import Optional, List, TYPE_CHECKING

import requests
import soundcloud
from bs4 import BeautifulSoup
from django.http.response import HttpResponse

from core.musiq import song_utils, musiq
from core.musiq.song_provider import SongProvider
from core.musiq.playlist_provider import PlaylistProvider

if TYPE_CHECKING:
    from core.musiq.song_utils import Metadata


class Soundcloud:
    """This class contains code for both the song and playlist provider"""

    _web_client: soundcloud.Client.Client = None  # type: ignore

    @staticmethod
    def _get_web_client() -> soundcloud.Client.Client:
        if Soundcloud._web_client is None:
            Soundcloud._web_client = soundcloud.Client(
                client_id="93e33e327fd8a9b77becd179652272e2"
            )
        return Soundcloud._web_client

    @property
    def web_client(self) -> soundcloud.Client.Client:
        """Returns the web client if it was already created.
        If not, it is created using the client-id from mopidy-soundcloud."""
        return Soundcloud._get_web_client()

    def get_search_suggestions(self, query: str) -> List[str]:
        """Returns a list of suggested items for the given query."""

        response = self.web_client.get(
            f"https://api-v2.soundcloud.com/search/queries", q=query
        )

        suggestions = [
            item.query
            for item in response.collection
            if not song_utils.is_forbidden(item.query)
        ]
        return suggestions


class SoundcloudSongProvider(SongProvider, Soundcloud):
    """This class handles songs from Soundcloud."""

    @staticmethod
    def get_id_from_external_url(url: str) -> str:
        track = Soundcloud._get_web_client().get("/resolve", url=url)
        return track.id

    @staticmethod
    def get_id_from_internal_url(url: str) -> str:
        """Returns the internal id based on the given url."""
        return url.split(".")[-1]

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "soundcloud"
        super().__init__(query, key)

        self.metadata: "Metadata" = {}
        self.external_url = None

    def check_cached(self) -> bool:
        # Soundcloud songs cannot be cached and have to be streamed everytime
        return False

    def check_available(self) -> bool:
        return self.gather_metadata()

    # track_info is of type mopidy.models.Track, but mopidy should not be a dependency, so no import
    def gather_metadata(self) -> bool:
        """Fetches metadata for this song's uri from Soundcloud."""
        if not self.id:
            results = self.web_client.get("/tracks", q=self.query, limit=20)

            # apply the filterlist from the settings
            for item in results:
                artist = item.user["username"]
                title = item.title
                if song_utils.is_forbidden(artist) or song_utils.is_forbidden(title):
                    continue
                result = item
                break
            else:
                # all tracks got filtered
                return False
            self.id = result.id
        else:
            result = self.web_client.get(f"tracks/{self.id}")
        self.metadata["artist"] = result.user["username"]
        self.metadata["title"] = result.title
        self.metadata["duration"] = result.duration / 1000
        self.metadata["internal_url"] = self.get_internal_url()
        self.metadata["external_url"] = result.permalink_url
        self.metadata["stream_url"] = None
        return True

    def get_metadata(self) -> "Metadata":
        if not self.metadata:
            self.gather_metadata()
        return self.metadata

    def _get_path(self) -> str:
        # soundcloud is not cached in the cache directory
        raise NotImplementedError()

    def get_internal_url(self) -> str:
        if not self.id:
            raise ValueError()
        return f"soundcloud:song.{self.id}"

    def get_external_url(self) -> str:
        if not self.id:
            raise ValueError()
        return self.get_metadata()["external_url"]

    def _get_related_urls(self) -> List[str]:
        response = requests.get(self.get_external_url() + "/recommended")

        soup = BeautifulSoup(response.text, "html.parser")

        # the first article is the current one
        articles = soup.select("article")[1:]
        urls = [
            "https://soundcloud.com" + article.select_one("a")["href"]
            for article in articles
        ]
        return urls

    def get_suggestion(self) -> str:
        return self._get_related_urls()[0]

    def request_radio(self, request_ip: str) -> HttpResponse:
        urls = self._get_related_urls()

        for external_url in urls:
            musiq.do_request_music(
                "",
                external_url,
                None,
                False,
                "soundcloud",
                archive=False,
                manually_requested=False,
            )

        return HttpResponse("queueing radio")


class SoundcloudPlaylistProvider(PlaylistProvider, Soundcloud):
    """This class handles Soundcloud Playlists."""

    @staticmethod
    def get_id_from_external_url(url: str) -> Optional[str]:
        if not url.startswith("https://soundcloud.com/"):
            return None
        playlist = Soundcloud._get_web_client().get("/resolve", url=url)
        return playlist.id

    def __init__(self, query: Optional[str], key: Optional[int]) -> None:
        self.type = "soundcloud"
        super().__init__(query, key)

    def search_id(self) -> Optional[str]:
        results = self.web_client.get("/playlists", q=self.query, limit=1)

        try:
            playlist = results[0]
        except IndexError:
            self.error = "No playlist found"
            return None

        list_id = playlist.id
        self.title = playlist.title

        return list_id

    def is_radio(self) -> bool:
        return False

    def fetch_metadata(self) -> bool:
        if self.title is None:
            result = self.web_client.get(f"/playlists/{self.id}")
            self.title = result.name

        tracks = self.web_client.get(f"/playlists/{self.id}/tracks")

        for track in tracks:
            self.urls.append(track.permalink_url)

        return True



"""This module provides app wide utility functions."""
import subprocess
from typing import List

from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponseForbidden, HttpResponse


def camelize(snake_dict: dict) -> dict:
    def camelize_str(snake: str) -> str:
        head, *tail = snake.split("_")
        return head + "".join(word.capitalize() for word in tail)

    return {camelize_str(k): v for k, v in snake_dict.items()}


def get_devices() -> List[str]:
    output = subprocess.check_output(
        "ip route show default".split(), universal_newlines=True
    )
    words = output.split()
    devices = []
    for cur, nex in zip(words, words[1:]):
        if cur == "dev":
            devices.append(nex)
    if not devices:
        raise ValueError("no devices found")
    return devices


def ip_of_device(device: str) -> str:
    output = subprocess.check_output(
        f"ip -4 a show dev {device}".split(), universal_newlines=True
    )
    ip = None
    for line in output.split("\n"):
        line = line.strip()
        if not line.startswith("inet"):
            continue
        ip = line.split()[1].split("/")[0]
        break
    if not ip:
        raise ValueError(f"ip not found for {device}")
    return ip


def broadcast_of_device(device: str) -> str:
    output = subprocess.check_output(
        f"ip -o -f inet addr show {device}".split(), universal_newlines=True
    )
    words = output.split()
    return words[5]


def service_installed(service: str) -> bool:
    if not service.endswith(".service"):
        service += ".service"
    out = subprocess.check_output(
        ["systemctl", "list-unit-files", service], universal_newlines=True
    )
    return len(out.splitlines()) > 3


def csrf_failure(_request: WSGIRequest, _reason: str = "") -> HttpResponse:
    return HttpResponseForbidden("Please reload")



"""This module handles playback of music."""

from __future__ import annotations

import datetime
import logging
import os
import random
import time
from contextlib import contextmanager
from threading import Event
from typing import Iterator, Optional

import requests

from core.celery import app
from django.conf import settings as conf
from django.db import transaction, connection
from django.utils import timezone
from mopidyapi.client import MopidyAPI
from mopidyapi.exceptions import MopidyError

from core import redis
import core.models as models
from core import user_manager
from core.lights import controller as lights_controller
from core.musiq import musiq, controller
from core.settings import storage, settings

queue_changed = redis.Event("queue_changed")
buzzer_stopped = redis.Event("buzzer_stopped")

queue = models.QueuedSong.objects

# this lock is released when restarting mopidy (which happens in another Thread)
player_lock = redis.lock("player_lock", thread_local=False)


def start() -> None:
    """Initializes this module by starting the playback and buzzer loop."""
    _handle_buzzer.delay()
    _loop.delay()


def set_playback_error(error: bool) -> None:
    """Sets the playback error. Updates musiq state when the state changes."""
    if redis.get("playback_error"):
        if not error:
            redis.set("playback_error", False)
            musiq.update_state()
            settings.update_state()
    else:
        if error:
            redis.set("playback_error", True)
            musiq.update_state()
            settings.update_state()


class Playback:
    """Class containing all playback related methods."""

    def __init__(self):
        # the celery worker needs its own player instance.
        # if we use a module-wide instance, methods can be used,
        # but events do not register correctly (probably due to thread boundary)
        if conf.TESTING:
            # to reduce the amount of created mopidy connections,
            # use the controller's instance during testing
            # this works because everything is run in a single process
            self.player = controller.player
        else:
            self.player: MopidyAPI = MopidyAPI(host=conf.MOPIDY_HOST)
        self.playback_started = Event()
        redis.set("playing", False)

        queue.delete_placeholders()

        with mopidy_command(important=True):
            self.player.playback.stop()
            self.player.tracklist.clear()
            # make songs disappear from tracklist after being played
            self.player.tracklist.set_consume(True)

        @self.player.on_event("track_playback_started")
        def _on_playback_started(_event) -> None:
            self.playback_started.set()

    def play_alarm(self, interrupt=False) -> None:
        """Play the alarm sound. If specified, interrupts the currently playing song."""
        redis.set("alarm_playing", True)
        lights_controller.alarm_started()
        self.playback_started.clear()

        with mopidy_command(important=True):
            # interrupt the current song if its playing
            if interrupt:
                self.player.tracklist.clear()
            self.player.tracklist.add(
                uris=[
                    "file://" + os.path.join(conf.BASE_DIR, "config/sounds/alarm.m4a")
                ]
            )
            self.player.playback.play()
        self.playback_started.wait(timeout=1)

        musiq.update_state()
        self.wait_until_song_end()

        lights_controller.alarm_stopped()
        redis.set("alarm_playing", False)

        if not interrupt:
            # if no song immediately continues playing, a manual state update is needed
            musiq.update_state()

    def wait_until_song_end(self) -> bool:
        """Wait until the song is over.
        Returns True when finished without errors, False otherwise."""
        # This is the event based approach. Unfortunately too error-prone.
        # If mopidy crashes/restarts for example, no track_playback_ended event is sent
        # playback_ended.wait()
        error = False
        while True:
            with mopidy_command() as allowed:
                if allowed:
                    try:
                        if self.player.playback.get_state() == "stopped":
                            break
                    except (requests.exceptions.ConnectionError, MopidyError):
                        # error during state get, skip until reconnected
                        error = True
            time.sleep(0.1)
            if redis.get("stop_playback_loop"):
                # in order to stop the playback thread, return False, making the main loop restart.
                # it will check this variable again and terminate itself.
                return False
            if redis.get("alarm_requested"):
                redis.set("alarm_requested", False)
                self.play_alarm(interrupt=True)
                # the current song was interrupted and needs to be resumed at the correct position
                # returning False will notify the main loop about this interruption,
                # making it restart the song correctly
                current_song = models.CurrentSong.objects.get()
                # we don't want the song to skip over the time when the alarm was playing
                # thus, we offset the creation date of the current song by the length of the alarm
                # Warning: if this duration does not fit the duration of the actual alarm,
                # Raveberry's internal state gets desynced and weird errors happen
                current_song.created += datetime.timedelta(
                    seconds=musiq.get_alarm_metadata()["duration"]
                )
                current_song.save()

                return False
        return not error

    def loop(self) -> None:
        """The main loop of the player.
        Takes a song from the queue and plays it until it is finished."""
        while True:

            if redis.get("stop_playback_loop"):
                break

            catch_up = None
            self.playback_started.clear()

            if models.CurrentSong.objects.exists():
                # recover interrupted song from database
                current_song = models.CurrentSong.objects.get()

                # continue with the current song (approximately) where we last left
                if storage.get("paused"):
                    catch_up = round(
                        (
                            current_song.last_paused - current_song.created
                        ).total_seconds()
                        * 1000
                    )
                else:
                    catch_up = round(
                        (timezone.now() - current_song.created).total_seconds() * 1000
                    )
                if catch_up > current_song.duration * 1000:
                    catch_up = -1
            else:
                if queue.count() == 0:
                    queue_changed.wait()
                    queue_changed.clear()

                    # restart the loop to check again whether songs are available
                    # in case of a false wakeup this causes as to wait again
                    continue

                # select the next song depending on settings
                song: Optional[models.QueuedSong]
                if storage.get("voting_enabled"):
                    with transaction.atomic():
                        song = queue.confirmed().order_by("-votes", "index")[0]
                        song_id = song.id
                        queue.remove(song.id)
                elif storage.get("shuffle"):
                    confirmed = queue.confirmed()
                    index = random.randint(0, confirmed.count() - 1)
                    song_id = confirmed[index].id
                    song = queue.remove(song_id)
                else:
                    # move the first song in the queue into the current song
                    song_id, song = queue.dequeue()

                if song is None:
                    # either the semaphore didn't match up with the actual count
                    # of songs in the queue or a race condition occured
                    logging.warning("dequeued on empty list")
                    continue

                if song.internal_url == "alarm":
                    self.play_alarm()
                    continue

                # stop backup stream.
                # when the dequeued song starts playing, the backup stream playback is stopped
                redis.set("backup_playing", False)

                current_song = models.CurrentSong.objects.create(
                    queue_key=song_id,
                    manually_requested=song.manually_requested,
                    votes=song.votes,
                    internal_url=song.internal_url,
                    external_url=song.external_url,
                    stream_url=song.stream_url,
                    artist=song.artist,
                    title=song.title,
                    duration=song.duration,
                )

                handle_autoplay()

                try:
                    archived_song = models.ArchivedSong.objects.get(
                        url=current_song.external_url
                    )
                    votes: Optional[int]
                    if storage.get("voting_enabled"):
                        votes = current_song.votes
                    else:
                        votes = None
                    if storage.get("logging_enabled"):
                        models.PlayLog.objects.create(
                            song=archived_song,
                            manually_requested=current_song.manually_requested,
                            votes=votes,
                        )
                except (
                    models.ArchivedSong.DoesNotExist,
                    models.ArchivedSong.MultipleObjectsReturned,
                ):
                    pass

            with mopidy_command(important=True):
                # after a restart consume may be set to False again, so make sure it is on
                self.player.tracklist.clear()
                self.player.tracklist.set_consume(True)
                self.player.tracklist.add(uris=[current_song.internal_url])
                # temporarily mute mopidy in case we need to seek but mopidy does not react directly
                # this allows us to seek first and then unmute, preventing audible skips
                volume = self.player.mixer.get_volume()
                if catch_up is not None and catch_up >= 0:
                    self.player.mixer.set_volume(0)
                # mopidy can only seek when the song is playing
                # also we do not continue without the playing state properly set.
                # otherwise waiting might exit before the song started
                self.player.playback.play()
                if not self.playback_started.wait(timeout=1):
                    # mopidy did not acknowledge that it started the song
                    # to make sure it is not in an error state,
                    # restart the loop and retry to start the song
                    # also prevents "queue-eating" bug,
                    # where mopidy in a failed state would refuse to play any song,
                    # but raveberry keeps on sending songs from the queue
                    logging.warning("playback_started event did not trigger")
                    set_playback_error(True)
                    self.player.mixer.set_volume(volume)
                    continue
                set_playback_error(False)
                if catch_up is not None and catch_up >= 0:
                    self.player.playback.seek(catch_up)
                    if storage.get("paused"):
                        self.player.playback.pause()
                    self.player.mixer.set_volume(volume)
            redis.set("playing", True)

            # needs some more testing but could prevent "eating the queue" bug
            # if not started_playing and not settings.DOCKER:
            #    logging.error(
            #        "Mopidy did not start playing the song in the timeout. Error assumed.\n"
            #        "Usually faulty outputs are the reason, config is reset to local output.\n"
            #        "Consult mopidy's log for more information.\n"
            #    )
            #    # reset to pulse device 0
            #    core.settings.sound._set_output("0")
            #    # recover the current song by restarting the loop
            #    continue

            musiq.update_state()

            # don't wait for the song to end if catch_up is negative (=the song should be skipped)
            if catch_up is None or catch_up >= 0:
                if not self.wait_until_song_end():
                    # there was an error while waiting for the song to end
                    # This happens when we could not connect to mopidy (ConnectionError)
                    # or when an interrupting alarm was initiated.
                    # we do not delete the current song but recover its state by restarting the loop
                    storage.set("paused", False)
                    redis.set("playing", False)
                    continue
            # Allowing new songs to start playing while paused introduces many edge cases
            # Instead of dealing with them, always start playback when skipping a paused song
            storage.set("paused", False)
            redis.set("playing", False)

            current_song.delete()

            if storage.get("repeat"):
                queue.enqueue(
                    {
                        "artist": current_song.artist,
                        "title": current_song.title,
                        "duration": current_song.duration,
                        "internal_url": current_song.internal_url,
                        "external_url": current_song.external_url,
                        "stream_url": current_song.stream_url,
                    },
                    False,
                )
                queue_changed.set()

            if user_manager.partymode_enabled() and random.random() < storage.get(
                "alarm_probability"
            ):
                self.play_alarm()

            if not queue.exists() and storage.get("backup_stream"):
                redis.set("backup_playing", True)
                # play backup stream
                self.player.tracklist.add(uris=[storage.get("backup_stream")])
                self.player.playback.play()

            musiq.update_state()


@app.task
def _loop() -> None:
    playback = Playback()
    playback.loop()
    connection.close()


@app.task
def _handle_buzzer() -> None:
    try:
        import gpiozero
    except ModuleNotFoundError:
        return
    buzzer = gpiozero.Button(16)
    last_press = timezone.now() - datetime.timedelta(
        seconds=storage.get("buzzer_cooldown")
    )

    def on_press():
        nonlocal last_press

        # do not allow the buzzer to be pressed too frequently
        if (timezone.now() - last_press).total_seconds() < storage.get(
            "buzzer_cooldown"
        ):
            logging.warning("buzzer pressed too quickly")
            return
        # do not allow an alarm to be triggered while one is already playing
        # or when an alarm is currently in the process of being played
        if redis.get("alarm_playing") or redis.get("alarm_requested"):
            logging.warning("last buzzer alarm not yet finished")
            return
        last_press = timezone.now()

        trigger_alarm()

    buzzer.when_pressed = on_press

    # wait until this task is told to exit
    buzzer_stopped.wait()


def trigger_alarm() -> None:
    """Initiate an alarm."""
    if redis.get("playing"):
        # if a song is currently playing, inform the loop waiting for the song to end
        # about this alarm. It will interrupt the current song and play the alarm
        redis.set("alarm_requested", True)
    else:
        # insert a special queue song to wake up the main loop and make it play the alarm
        queue.enqueue(musiq.get_alarm_metadata(), True)
        queue_changed.set()


def handle_autoplay(url: Optional[str] = None) -> None:
    """Checks whether to add a song by autoplay and does so if necessary.
    :param url: if given, this url is used to find the next autoplayed song.
    Otherwise, the current song is used."""
    if storage.get("autoplay") and models.QueuedSong.objects.count() == 0:
        if url is None:
            # if no url was specified, use the one of the current song
            try:
                current_song = models.CurrentSong.objects.get()
                url = current_song.external_url
            except (
                models.CurrentSong.DoesNotExist,
                models.CurrentSong.MultipleObjectsReturned,
            ):
                return

        from core.musiq.song_provider import SongProvider

        provider = SongProvider.create(external_url=url)
        try:
            suggestion = provider.get_suggestion()
            # The player loop is not restarted after error automatically.
            # As this function can raise several exceptions (it might do networking)
            # we catch every exception to make sure the loop keeps running
        except Exception as e:  # pylint: disable=broad-except
            logging.exception("error during suggestions for %s: %s", url, e)
        else:
            musiq.do_request_music(
                "",
                suggestion,
                None,
                False,
                provider.type,
                archive=False,
                manually_requested=False,
            )


@contextmanager
def mopidy_command(important: bool = False) -> Iterator[bool]:
    """A context that should be used around every mopidy command used.
    Makes sure that commands occur sequentially, as mopidy can not handle parallel inputs.
    Use it like this:
    with mopidy_command() as allowed:
        if allowed:
            # mopidy command
    :param important: If True, wait until the lock is released.
    If not, return 'False' after a timeout."""
    timeout = 3
    if important:
        timeout = None
    if player_lock.acquire(blocking_timeout=timeout):
        yield True
        player_lock.release()
    else:
        logging.warning("mopidy command could not be executed")
        set_playback_error(True)
        yield False


def stop() -> None:
    """Stops the playback main loop, only used for tests."""
    redis.set("stop_playback_loop", True)
    queue_changed.set()



# Generated by Django 2.2.20 on 2021-04-27 11:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("core", "0010_tag_active")]

    operations = [
        migrations.AddField(
            model_name="currentsong",
            name="stream_url",
            field=models.CharField(blank=True, max_length=2000),
        ),
        migrations.AddField(
            model_name="queuedsong",
            name="stream_url",
            field=models.CharField(blank=True, max_length=2000),
        ),
    ]



# Generated by Django 2.2.10 on 2020-05-14 10:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("core", "0007_remove_playlistentry_song")]

    operations = [
        migrations.AlterField(
            model_name="archivedplaylist",
            name="list_id",
            field=models.CharField(max_length=2000),
        ),
        migrations.AlterField(
            model_name="archivedsong",
            name="url",
            field=models.CharField(max_length=2000, unique=True),
        ),
        migrations.AlterField(
            model_name="currentsong",
            name="external_url",
            field=models.CharField(blank=True, max_length=2000),
        ),
        migrations.AlterField(
            model_name="currentsong",
            name="internal_url",
            field=models.CharField(max_length=2000),
        ),
        migrations.AlterField(
            model_name="playlistentry",
            name="url",
            field=models.CharField(max_length=2000),
        ),
        migrations.AlterField(
            model_name="queuedsong",
            name="external_url",
            field=models.CharField(blank=True, max_length=2000),
        ),
        migrations.AlterField(
            model_name="queuedsong",
            name="internal_url",
            field=models.CharField(max_length=2000),
        ),
    ]



# Generated by Django 2.2.9 on 2020-01-30 00:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("core", "0005_auto_20190913_1730")]

    operations = [
        migrations.RenameField(
            model_name="currentsong", old_name="url", new_name="internal_url"
        ),
        migrations.RenameField(
            model_name="queuedsong", old_name="url", new_name="internal_url"
        ),
        migrations.AddField(
            model_name="currentsong",
            name="external_url",
            field=models.CharField(blank=True, max_length=200),
        ),
        migrations.AddField(
            model_name="queuedsong",
            name="external_url",
            field=models.CharField(blank=True, max_length=200),
        ),
        migrations.RemoveField(model_name="currentsong", name="location"),
        migrations.AlterField(
            model_name="currentsong", name="duration", field=models.IntegerField()
        ),
        migrations.AlterField(
            model_name="queuedsong", name="duration", field=models.IntegerField()
        ),
    ]



"""This module handles realtime communication via websockets."""
import json
from typing import Dict, Any

from asgiref.sync import async_to_sync
from channels.generic.websocket import WebsocketConsumer
from channels.layers import get_channel_layer
from django.core.handlers.wsgi import WSGIRequest
from django.http import JsonResponse


def send_state(state: Dict[str, Any]) -> None:
    """Sends the given dictionary as a state update to all connected clients."""
    data = {"type": "state_update", "state": state}
    channel_layer = get_channel_layer()
    async_to_sync(channel_layer.group_send)("state", data)


def get_state(_request: WSGIRequest, module) -> JsonResponse:
    state = module.state_dict()
    return JsonResponse(state)


class StateConsumer(WebsocketConsumer):
    """Handles connections with websocket clients."""

    def connect(self) -> None:
        async_to_sync(self.channel_layer.group_add)("state", self.channel_name)
        self.accept()

    def disconnect(self, code: int) -> None:
        async_to_sync(self.channel_layer.group_discard)("state", self.channel_name)

    def receive(self, text_data: str = None, bytes_data: bytes = None) -> None:
        pass

    def state_update(self, event: Dict[str, Any]):
        """Receives a message from the room group and sends it back to the websocket."""
        self.send(text_data=json.dumps(event["state"]))



"""This module handles all controls that change the playback."""

from __future__ import annotations

import subprocess
import datetime
from functools import wraps
from typing import Callable

from django.conf import settings as conf
from django.core.handlers.wsgi import WSGIRequest
from django.db import transaction
from django.db.models import F
from django.http import HttpResponseForbidden
from django.http.response import HttpResponse, HttpResponseBadRequest
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt
from mopidyapi import MopidyAPI

import core.models as models
from core import user_manager, redis
from core.musiq import playback, musiq
from core.settings import storage

player: MopidyAPI = None


def control(func: Callable) -> Callable:
    """A decorator for functions that control the playback.
    Every control changes the views state and returns an empty response.
    At least mod privilege is required during voting."""

    def _decorator(request: WSGIRequest) -> HttpResponse:
        if storage.get("voting_enabled") and not user_manager.has_controls(
            request.user
        ):
            return HttpResponseForbidden()
        response = func(request)
        musiq.update_state()
        if response is not None:
            return response
        return HttpResponse()

    return wraps(func)(_decorator)


SEEK_DISTANCE = 10


def start() -> None:
    """Initializes this module by restoring the volume."""
    global player
    player = MopidyAPI(host=conf.MOPIDY_HOST)
    volume = storage.get("volume")
    _set_volume(volume)


@control
def restart(_request: WSGIRequest) -> None:
    """Restarts the current song from the beginning."""
    with playback.mopidy_command() as allowed:
        if allowed:
            player.playback.seek(0)
    try:
        current_song = models.CurrentSong.objects.get()
        current_song.created = timezone.now()
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass


@control
def seek_backward(_request: WSGIRequest) -> None:
    """Jumps back in the current song."""
    with playback.mopidy_command() as allowed:
        if allowed:
            current_position = player.playback.get_time_position()
            player.playback.seek(current_position - SEEK_DISTANCE * 1000)
    try:
        current_song = models.CurrentSong.objects.get()
        now = timezone.now()
        current_song.created += datetime.timedelta(seconds=SEEK_DISTANCE)
        current_song.created = min(current_song.created, now)
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass


@control
def play(_request: WSGIRequest) -> None:
    """Resumes the current song if it is paused.
    No-op if already playing."""
    with playback.mopidy_command() as allowed:
        if allowed:
            player.playback.play()
    try:
        # move the creation timestamp into the future for the duration of the pause
        # this ensures that the progress calculation (starting from created) is correct
        current_song = models.CurrentSong.objects.get()
        now = timezone.now()
        pause_duration = (now - current_song.last_paused).total_seconds()
        current_song.created += datetime.timedelta(seconds=pause_duration)
        current_song.created = min(current_song.created, now)
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass
    storage.set("paused", False)


@control
def pause(_request: WSGIRequest) -> None:
    """Pauses the current song if it is playing.
    No-op if already paused."""
    with playback.mopidy_command() as allowed:
        if allowed:
            player.playback.pause()
    try:
        current_song = models.CurrentSong.objects.get()
        current_song.last_paused = timezone.now()
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass
    storage.set("paused", True)


@control
def seek_forward(_request: WSGIRequest) -> None:
    """Jumps forward in the current song."""
    with playback.mopidy_command() as allowed:
        if allowed:
            current_position = player.playback.get_time_position()
            player.playback.seek(current_position + SEEK_DISTANCE * 1000)
    try:
        current_song = models.CurrentSong.objects.get()
        current_song.created -= datetime.timedelta(seconds=SEEK_DISTANCE)
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass


@control
def skip(_request: WSGIRequest) -> None:
    """Skips the current song and continues with the next one."""
    with playback.mopidy_command() as allowed:
        if allowed:
            redis.set("backup_playing", False)
            player.playback.next()


@control
def set_shuffle(request: WSGIRequest) -> None:
    """Enables or disables shuffle depending on the given value.
    If enabled, a random song in the queue is chosen as the next one.
    If not, the first one is chosen."""
    enabled = request.POST.get("value") == "true"
    storage.set("shuffle", enabled)


@control
def set_repeat(request: WSGIRequest) -> None:
    """Enables or disables repeat depending on the given value.
    If enabled, a song is enqueued again after it finished playing."""
    enabled = request.POST.get("value") == "true"
    storage.set("repeat", enabled)


@control
def set_autoplay(request: WSGIRequest) -> None:
    """Enables or disables autoplay depending on the given value.
    If enabled and the current song is the last one,
    a new song is enqueued, based on the current one."""
    enabled = request.POST.get("value") == "true"
    storage.set("autoplay", enabled)
    playback.handle_autoplay()


def _set_volume(volume) -> None:
    try:
        # Try to set the volume via the pulse server.
        # This is faster and does not impact visualization
        subprocess.run(
            f"pactl set-sink-volume @DEFAULT_SINK@ {round(volume*100)}%".split(),
            env={"PULSE_SERVER": "127.0.0.1"},
            check=True,
        )
    except (FileNotFoundError, subprocess.CalledProcessError):
        # pulse is not installed or there is no server running.
        # change mopidy's volume
        with playback.mopidy_command() as allowed:
            if allowed:
                player.mixer.set_volume(round(volume * 100))


@control
def set_volume(request: WSGIRequest) -> None:
    """Sets the playback volume.
    value has to be a float between 0 and 1."""
    volume = float(request.POST.get("value"))  # type: ignore
    _set_volume(volume)
    storage.set("volume", volume)


@control
def shuffle_all(request: WSGIRequest) -> HttpResponse:
    """Shuffles the queue. Only admin is permitted to do this."""
    if not user_manager.is_admin(request.user):
        return HttpResponseForbidden()
    playback.queue.shuffle()
    return HttpResponse()


@control
def remove_all(request: WSGIRequest) -> HttpResponse:
    """Empties the queue. Only admin is permitted to do this."""
    if not user_manager.is_admin(request.user):
        return HttpResponseForbidden()
    with playback.mopidy_command() as allowed:
        if allowed:
            with transaction.atomic():
                playback.queue.all().delete()
    return HttpResponse()


@control
def prioritize(request: WSGIRequest) -> HttpResponse:
    """Prioritizes song by making it the first one in the queue."""
    key = request.POST.get("key")
    if key is None:
        return HttpResponseBadRequest()
    ikey = int(key)
    playback.queue.prioritize(ikey)
    return HttpResponse()


@control
def remove(request: WSGIRequest) -> HttpResponse:
    """Removes a song identified by the given key from the queue."""
    key = request.POST.get("key")
    if key is None:
        return HttpResponseBadRequest()
    ikey = int(key)
    try:
        removed = playback.queue.remove(ikey)
        # if we removed a song and it was added by autoplay,
        # we want it to be the new basis for autoplay
        if not removed.manually_requested:
            playback.handle_autoplay(removed.external_url or removed.title)
        else:
            playback.handle_autoplay()
    except models.QueuedSong.DoesNotExist:
        return HttpResponseBadRequest("song does not exist")
    return HttpResponse()


@control
def reorder(request: WSGIRequest) -> HttpResponse:
    """Reorders the queue.
    The song specified by element is inserted between prev and next."""
    prev_key = request.POST.get("prev")
    cur_key = request.POST.get("element")
    next_key = request.POST.get("next")
    if not cur_key:
        return HttpResponseBadRequest()
    if not prev_key:
        iprev_key = None
    else:
        iprev_key = int(prev_key)
    icur_key = int(cur_key)
    if not next_key:
        inext_key = None
    else:
        inext_key = int(next_key)
    try:
        playback.queue.reorder(iprev_key, icur_key, inext_key)
    except ValueError:
        return HttpResponseBadRequest("request on old state")
    return HttpResponse()


@csrf_exempt
def vote_up(request: WSGIRequest) -> HttpResponse:
    """Increases the vote-count of the given song by one."""
    key = request.POST.get("key")
    if key is None:
        return HttpResponseBadRequest()
    ikey = int(key)

    models.CurrentSong.objects.filter(queue_key=ikey).update(votes=F("votes") + 1)
    playback.queue.vote_up(ikey)
    musiq.update_state()
    return HttpResponse()


@csrf_exempt
def vote_down(request: WSGIRequest) -> HttpResponse:
    """Decreases the vote-count of the given song by one.
    If a song receives too many downvotes, it is removed."""
    key = request.POST.get("key")
    if key is None:
        return HttpResponseBadRequest()
    ikey = int(key)

    models.CurrentSong.objects.filter(queue_key=ikey).update(votes=F("votes") - 1)
    try:
        current_song = models.CurrentSong.objects.get()
        if current_song.queue_key == ikey and current_song.votes <= -storage.get(
            "downvotes_to_kick"
        ):
            with playback.mopidy_command() as allowed:
                if allowed:
                    player.playback.next()
    except models.CurrentSong.DoesNotExist:
        pass

    removed = playback.queue.vote_down(ikey, -storage.get("downvotes_to_kick"))
    # if we removed a song by voting, and it was added by autoplay,
    # we want it to be the new basis for autoplay
    if removed is not None:
        if not removed.manually_requested:
            playback.handle_autoplay(removed.external_url or removed.title)
        else:
            playback.handle_autoplay()
    musiq.update_state()
    return HttpResponse()



import setuptools

with open("raveberry/VERSION") as f:
    version = f.read().strip()

with open("raveberry/README.md") as f:
    long_description = f.read()

with open("raveberry/requirements/common.txt") as f:
    run_packages = f.read().splitlines()
with open("raveberry/requirements/youtube.txt") as f:
    run_packages.extend(f.read().splitlines())

with open("raveberry/requirements/install.txt") as f:
    install_packages = f.read().splitlines()

with open("raveberry/requirements/screenvis.txt") as f:
    screenvis_packages = f.read().splitlines()

setuptools.setup(
    name="raveberry",
    version=version,
    author="Jonathan Hacker",
    author_email="raveberry@jhacker.de",
    description="A multi-user music server with a focus on participation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/raveberry/raveberry",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Django :: 2.2",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python :: 3",
    ],
    packages=["raveberry"],
    include_package_data=True,
    python_requires=">=3.7",
    extras_require={
        "install": install_packages,
        "run": run_packages,
        "screenvis": screenvis_packages,
    },
    scripts=["raveberry/bin/raveberry"],
)



# Generated by Django 2.2.13 on 2020-07-26 20:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("core", "0008_auto_20200514_1213")]

    operations = [migrations.DeleteModel(name="Pad")]



"""This module handles all settings related to the local library."""

from __future__ import annotations

import logging
import os
import time

from django.conf import settings as conf
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import JsonResponse
from mutagen import MutagenError

import core.musiq.song_utils as song_utils
from core import redis
from core.celery import app
from core.models import ArchivedSong, ArchivedPlaylist, PlaylistEntry
from core.settings import settings
from core.settings.settings import control


def get_library_path() -> str:
    """Returns the absolute path of the local library."""
    return os.path.abspath(os.path.join(conf.SONGS_CACHE_DIR, "local_library"))


@control
def list_subdirectories(request: WSGIRequest) -> HttpResponse:
    """Returns a list of all subdirectories for the given path."""
    path = request.GET.get("path")
    if path is None:
        return HttpResponseBadRequest("path was not supplied.")
    basedir, subdirpart = os.path.split(path)
    if path == "":
        suggestions = ["/"]
    elif os.path.isdir(basedir):
        suggestions = [
            os.path.join(basedir, subdir + "/")
            for subdir in next(os.walk(basedir))[1]
            if subdir.lower().startswith(subdirpart.lower())
        ]
        suggestions.sort()
    else:
        suggestions = ["not a valid directory"]
    if not suggestions:
        suggestions = ["not a valid directory"]
    return JsonResponse(suggestions, safe=False)


def _set_scan_progress(scan_progress: str) -> None:
    redis.set("library_scan_progress", scan_progress)
    settings.update_state()


@control
def scan_library(request: WSGIRequest) -> HttpResponse:
    """Scan the folder at the given path and add all its sound files to the database."""
    library_path = request.POST.get("library_path")
    if library_path is None:
        return HttpResponseBadRequest("library path was not supplied.")

    if not os.path.isdir(library_path):
        return HttpResponseBadRequest("not a directory")
    library_path = os.path.abspath(library_path)

    _set_scan_progress("0 / 0 / 0")

    _scan_library.delay(library_path)

    return HttpResponse(f"started scanning in {library_path}. This could take a while")


@app.task
def _scan_library(library_path: str) -> None:
    scan_start = time.time()
    last_update = scan_start
    update_frequency = 0.5
    filecount = 0
    for (dirpath, _, filenames) in os.walk(library_path):
        now = time.time()
        if now - last_update > update_frequency:
            last_update = now
            _set_scan_progress(f"{filecount} / 0 / 0")
        if os.path.abspath(dirpath) == os.path.abspath(conf.SONGS_CACHE_DIR):
            # do not add files handled by raveberry as local files
            continue
        filecount += len(filenames)

    library_link = os.path.join(conf.SONGS_CACHE_DIR, "local_library")
    try:
        os.remove(library_link)
    except FileNotFoundError:
        pass
    os.symlink(library_path, library_link)

    logging.info("started scanning in %s", library_path)

    _set_scan_progress(f"{filecount} / 0 / 0")

    files_scanned = 0
    files_added = 0
    for (dirpath, _, filenames) in os.walk(library_path):
        if os.path.abspath(dirpath) == os.path.abspath(conf.SONGS_CACHE_DIR):
            # do not add files handled by raveberry as local files
            continue
        now = time.time()
        if now - last_update > update_frequency:
            last_update = now
            _set_scan_progress(f"{filecount} / {files_scanned} / {files_added}")
        for filename in filenames:
            files_scanned += 1
            path = os.path.join(dirpath, filename)
            try:
                metadata = song_utils.get_metadata(path)
            except (ValueError, MutagenError):
                # the given file could not be parsed and will not be added to the database
                pass
            else:
                library_relative_path = path[len(library_path) + 1 :]
                external_url = os.path.join("local_library", library_relative_path)
                if not ArchivedSong.objects.filter(url=external_url).exists():
                    files_added += 1
                    ArchivedSong.objects.create(
                        url=external_url,
                        artist=metadata["artist"],
                        title=metadata["title"],
                        counter=0,
                    )

    assert files_scanned == filecount
    _set_scan_progress(f"{filecount} / {files_scanned} / {files_added}")

    logging.info("done scanning in %s", library_path)


@control
def create_playlists(_request: WSGIRequest) -> HttpResponse:
    """Create a playlist for every folder in the library."""
    library_link = os.path.join(conf.SONGS_CACHE_DIR, "local_library")
    if not os.path.islink(library_link):
        return HttpResponseBadRequest("No library set")

    _set_scan_progress(f"0 / 0 / 0")

    _create_playlists.delay()

    return HttpResponse(f"started creating playlists. This could take a while")


@app.task
def _create_playlists() -> None:
    local_files = ArchivedSong.objects.filter(url__startswith="local_library").count()

    library_link = os.path.join(conf.SONGS_CACHE_DIR, "local_library")
    library_path = os.path.abspath(library_link)

    logging.info("started creating playlists in %s", library_path)

    _set_scan_progress(f"{local_files} / 0 / 0")

    scan_start = time.time()
    last_update = scan_start
    update_frequency = 0.5
    files_processed = 0
    files_added = 0
    for (dirpath, _, filenames) in os.walk(library_path):
        now = time.time()
        if now - last_update > update_frequency:
            last_update = now
            _set_scan_progress(f"{local_files} / {files_processed} / {files_added}")

        song_urls = []
        # unfortunately there is no way to access track numbers accross different file types
        # so we have to add songs to playlists alphabetically
        for filename in sorted(filenames):
            path = os.path.join(dirpath, filename)
            library_relative_path = path[len(library_path) + 1 :]
            external_url = os.path.join("local_library", library_relative_path)
            if ArchivedSong.objects.filter(url=external_url).exists():
                files_processed += 1
                song_urls.append(external_url)

        if not song_urls:
            continue

        playlist_id = os.path.join("local_library", dirpath[len(library_path) + 1 :])
        playlist_title = os.path.split(dirpath)[1]
        playlist, created = ArchivedPlaylist.objects.get_or_create(
            list_id=playlist_id, title=playlist_title, counter=0
        )
        if not created:
            # this playlist already exists, skip
            continue

        song_index = 0
        for external_url in song_urls:
            PlaylistEntry.objects.create(
                playlist=playlist, index=song_index, url=external_url
            )
            files_added += 1
            song_index += 1

    _set_scan_progress(f"{local_files} / {files_processed} / {files_added}")



from typing import Union, Any, Tuple

from ast import literal_eval

import core.models as models

# maps key to default and type of value
defaults = {
    # basic settings
    "voting_enabled": False,
    "new_music_only": False,
    "logging_enabled": True,
    "hashtags_active": True,
    "embed_stream": False,
    "dynamic_embedded_stream": False,
    "online_suggestions": True,
    "number_of_suggestions": 20,
    "people_to_party": 3,
    "alarm_probability": 0.0,
    "buzzer_cooldown": 60,
    "downvotes_to_kick": 2,
    "max_download_size": 0.0,
    "max_playlist_items": 10,
    "max_queue_length": 0,
    "additional_keywords": "",
    "forbidden_keywords": "",
    # platforms
    "local_enabled": True,
    "youtube_enabled": True,
    "youtube_suggestions": 2,
    "spotify_enabled": False,
    "spotify_suggestions": 2,
    "spotify_username": "",
    "spotify_password": "",
    "spotify_client_id": "",
    "spotify_client_secret": "",
    "soundcloud_enabled": False,
    "soundcloud_suggestions": 2,
    "soundcloud_auth_token": "",
    "jamendo_enabled": False,
    "jamendo_suggestions": 2,
    "jamendo_client_id": "",
    # sound
    "feed_cava": True,
    "output": "",
    "backup_stream": "",
    # playback
    "paused": False,
    "volume": 1.0,
    "shuffle": False,
    "repeat": False,
    "autoplay": False,
    # lights
    "fixed_color": (0, 0, 0),
    "program_speed": 0.5,
    "wled_led_count": 10,
    "wled_ip": "",
    "wled_port": 21324,
    # the concise, but not much shorter version:
    # **{
    #    k: v
    #    for k, v in list(
    #        chain.from_iterable(
    #            [
    #                (f"{device}_brightness", 1.0),
    #                (f"{device}_monochrome", False),
    #                (f"{device}_program", "Disabled"),
    #                (f"last_{device}_program", "Disabled"),
    #            ]
    #            for device in ["ring", "strip", "wled", "screen"]
    #        )
    #    )
    # },
    "ring_brightness": 1.0,
    "ring_monochrome": False,
    "ring_program": "Disabled",
    "last_ring_program": "Disabled",
    "strip_brightness": 1.0,
    "strip_monochrome": False,
    "strip_program": "Disabled",
    "last_strip_program": "Disabled",
    "wled_brightness": 1.0,
    "wled_monochrome": False,
    "wled_program": "Disabled",
    "last_wled_program": "Disabled",
    "screen_brightness": 1.0,
    "screen_monochrome": False,
    "screen_program": "Disabled",
    "last_screen_program": "Disabled",
}


def get(key: str) -> Union[bool, int, float, str, Tuple]:
    """This method returns the value for the given :param key:.
    Vaules of non-existing keys are set to their respective default value."""
    # values are stored as string in the database
    # cast the value to its respective type, defined by the default value, before returning it
    default = defaults[key]
    value = models.Setting.objects.get_or_create(
        key=key, defaults={"value": str(default)}
    )[0].value
    if type(default) in (str, int, float):
        return type(default)(value)
    if type(default) == bool:
        # bool("False") does not return False -> special case for bool
        return value == "True"
    elif type(default) == tuple:
        # evaluate the stored literal
        return literal_eval(value)


def set(key: str, value: Any) -> None:
    """This method sets the :param value: for the given :param key:."""
    default = defaults[key]
    setting = models.Setting.objects.get_or_create(
        key=key, defaults={"value": str(default)}
    )[0]
    setting.value = value
    setting.save()



# Generated by Django 2.2.19 on 2021-04-26 22:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("core", "0009_delete_pad")]

    operations = [
        migrations.AddField(
            model_name="tag",
            name="active",
            field=models.BooleanField(default=True),
            preserve_default=False,
        )
    ]



"""This module handles all settings related to wifi."""

from __future__ import annotations

import os
import subprocess

from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import JsonResponse

from core.settings.settings import control


@control
def available_ssids(_request: WSGIRequest) -> JsonResponse:
    """List all ssids that can currently be seen."""
    output = subprocess.check_output(
        ["sudo", "/usr/local/sbin/raveberry/list_available_ssids"]
    ).decode()
    ssids = output.split("\n")
    return JsonResponse(ssids[:-1], safe=False)


@control
def connect_to_wifi(request: WSGIRequest) -> HttpResponse:
    """Connect to a given ssid with the given password."""
    ssid = request.POST.get("ssid")
    password = request.POST.get("password")
    if ssid is None or password is None or ssid == "" or password == "":
        return HttpResponseBadRequest("Please provide both SSID and password")
    try:
        output = subprocess.check_output(
            ["sudo", "/usr/local/sbin/raveberry/connect_to_wifi", ssid, password]
        ).decode()
        return HttpResponse(output)
    except subprocess.CalledProcessError as e:
        output = e.output.decode()
        return HttpResponseBadRequest(output)


@control
def disable_homewifi(_request: WSGIRequest) -> None:
    """Disable homewifi function."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_homewifi"])


@control
def enable_homewifi(_request: WSGIRequest) -> None:
    """Enable homewifi function."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_homewifi"])


@control
def stored_ssids(_request: WSGIRequest) -> JsonResponse:
    """Return the list of ssids that this Raspberry Pi was connected to in the past."""
    output = subprocess.check_output(
        ["sudo", "/usr/local/sbin/raveberry/list_stored_ssids"]
    ).decode()
    ssids = output.split("\n")
    return JsonResponse(ssids[:-1], safe=False)


@control
def set_homewifi_ssid(request: WSGIRequest) -> HttpResponse:
    """Set the home network.
    The hotspot will not be created if connected to this wifi."""
    homewifi_ssid = request.POST.get("value")
    if homewifi_ssid is None:
        return HttpResponseBadRequest("homewifi ssid was not supplied.")
    with open(os.path.join(settings.BASE_DIR, "config/homewifi"), "w+") as f:
        f.write(homewifi_ssid)
    return HttpResponse()



"""This module contains all url endpoints and maps them to their corresponding functions."""
import inspect

from typing import Any, List

from django.core.handlers.wsgi import WSGIRequest
from django.urls import include, URLPattern
from django.urls import path
from django.views.generic import RedirectView

import core.api as api
import core.network_info as network_info
import core.base as base
import core.lights.lights as lights
import core.lights.controller as lights_controller
import core.musiq.musiq as musiq
import core.musiq.controller as musiq_controller
import core.musiq.suggestions as suggestions
import core.settings.settings as settings
import core.settings.analysis as analysis
import core.settings.basic as basic
import core.settings.library as library
import core.settings.platforms as platforms
import core.settings.sound as sound
import core.settings.system as system
import core.settings.wifi as wifi
from core import state_handler

urlpatterns = [
    path("", RedirectView.as_view(pattern_name="musiq", permanent=False), name="base"),
    path("musiq/", musiq.index, name="musiq"),
    path("lights/", lights.index, name="lights"),
    path("stream/", base.no_stream, name="no-stream"),
    path("network_info/", network_info.index, name="network-info"),
    path("settings/", settings.index, name="settings"),
    path("accounts/", include("django.contrib.auth.urls")),
    path("login/", RedirectView.as_view(pattern_name="login", permanent=False)),
    path("logged_in/", base.logged_in, name="logged-in"),
    path("logout/", RedirectView.as_view(pattern_name="logout", permanent=False)),
    path(
        "api/",
        include(
            [
                path("version/", api.version, name="version"),
                path(
                    "musiq/",
                    include([path("post_song/", api.post_song, name="post-song")]),
                ),
            ]
        ),
    ),
]

urlmethods = [urlpattern.callback for urlpattern in urlpatterns]


def get_paths(objs: List[Any]) -> List[URLPattern]:
    """Iterates through the given objects and identifies all methods that serve http requests.
    A url pattern is generated for each of these methods,
    but only if no such path exists already in urlpatterns.
    Returns the list of url patterns."""
    paths = []
    for obj in objs:
        for name, method in inspect.getmembers(obj, inspect.isfunction):
            if name == "get_state":
                # the state url is an exception
                # it cannot be named after its method name, as every page has its own get_state
                continue
            if name.startswith("_"):
                # do not expose internal methods
                continue
            # use __annotations__ instead of get_type_hints, because the latter raised errors about
            # WSGIRequest not being defined in some configurations
            # string annotations are sufficient for our use case
            type_hints = method.__annotations__
            if "request" in type_hints:
                request_type = type_hints["request"]
            elif "_request" in type_hints:
                request_type = type_hints["_request"]
            else:
                continue
            if (
                request_type == "WSGIRequest" or issubclass(request_type, WSGIRequest)
            ) and method not in urlmethods:
                name = name.replace("_", "-")
                paths.append(path(name + "/", method, name=name))
    return paths


base_paths = get_paths([base])
musiq_paths = get_paths([musiq, musiq_controller, suggestions])
lights_paths = get_paths([lights_controller])
settings_paths = get_paths([basic, platforms, sound, wifi, library, analysis, system])

urlpatterns.append(
    path(
        "ajax/",
        include(
            [
                path("", include(base_paths)),
                path(
                    "musiq/state/",
                    state_handler.get_state,
                    {"module": musiq},
                    name="musiq-state",
                ),
                path("musiq/", include(musiq_paths)),
                path(
                    "lights/state/",
                    state_handler.get_state,
                    {"module": lights},
                    name="lights-state",
                ),
                path("lights/", include(lights_paths)),
                path(
                    "settings/state/",
                    state_handler.get_state,
                    {"module": settings},
                    name="settings-state",
                ),
                path("settings/", include(settings_paths)),
            ]
        ),
    )
)



"""This module contains the device superclass."""
from core import redis
from core.settings import storage


class Device:
    """A class representing a visualization device that Raveberry can control."""

    def __init__(self, manager, name) -> None:
        self.manager = manager
        self.name = name
        self.brightness = storage.get(f"{self.name}_brightness")
        self.monochrome = storage.get(f"{self.name}_monochrome")
        self.initialized = False
        redis.set(f"{self.name}_initialized", False)
        self.program = None

    def load_program(self) -> None:
        """Load and activate this device's program from the database."""
        program_name = storage.get(f"{self.name}_program")

        # only enable if the device is initialized
        if self.initialized:
            self.program = self.manager.all_programs[program_name]
        else:
            self.program = self.manager.disabled_program

        self.program.use()

    def clear(self) -> None:
        """Resets this device, clearing all visualization."""
        raise NotImplementedError()
